  /* utilities for generating script files for Damian's CCM program 
     (w) Sven Krueger Aug09 */
  
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define P(EX)  printf("%s:  %d\n",#EX,(int)(EX))
#define Pf(EX) printf("%s: %13.6e\n",#EX,(EX))  /* full precision float */
#define Pd(EX) printf("%s: %22.15e\n",#EX,(EX)) /* full precision double */
#define Ps(EX) printf("%s:  %s\n",#EX,(EX))

#define EE(EX1)          {printf("\n%s: "EX1" in %s() at L%d: ",__FILE__,__func__,__LINE__);}
#define E0(EX1,EX2)      {EE("error"); printf(EX1"\n",EX2); exit(1);}


#define _nearZero 1.0e-10
char term_descr[14][24] = { "",   "zz cos(phi)",             // term 1
                           "z+ -.5*sin(phi)", "z- -.5*sin(phi)", "+z .5*sin(phi)", "-z .5*sin(phi)", // term 2..5
                           "+- (cos(phi)+1)/4", "-+ (cos(phi)+1)/4", "++ (cos(phi)-1)/4", "-- (cos(phi)-1)/4", // term 6..9
                           "z",  "-",  "z^2","+"}; // term 10..13 


  /* print i-th nn of site */
void print_nn(int site, int i)
{ int j;
  for (j=0;j<3;j++) printf("%3d",bond[n_nn*(site-1)+i].nn[j]);
  j=3; printf(" %3d",bond[n_nn*(site-1)+i].nn[j]);
}

int round1( double x) 
{ 
  if (x>=0.0) return (int)(x + 0.5); 
  else        return (int)(x - 0.5);
}
  /* print a*pi/b, given x (=a/b) */
char *print_frac(double x, char *buf)
{ int i, a;
  for (i=1;i<20;i++) {
    a = round1(x*i);
    if (fabs(a-x*i)<1.0e-3) {
      if       (a==0) sprintf(buf,"0");
      else if ((a==1)&&(i==1)) sprintf(buf,"pi");
      else if ((a==1)&&(i!=1)) sprintf(buf,"pi/%d",i);
      else if ((a!=1)&&(i==1)) sprintf(buf,"%dpi",a);
      else sprintf(buf,"%dpi/%d",a,i);
      return(buf);
    }
  }
  sprintf(buf,"%6.4fpi",x);
  return(buf);
}
  /* print a double as %f with i digits */
char *print_float(double x, int i, char *buf)
{ char str1[100];
  sprintf(str1," %%%d.%df",i+4,i);
  sprintf(buf,str1,x);
  return(buf);
}
  /* -0.0 --> 0.0 */
double noNegZ(double x) {
  if (fabs(x)<_nearZero) return(0.0); 
  else return(x);
}
  /* rotate vector clockwise around z-axis (right-hand system) 
     == positive (anti-clockwise) rotation of coordinate system */
void rot(double phi,double *tr, char *comment)
{ char buf[100]; double a11,a12,a21,a22;
  a11=noNegZ( cos(phi)); a12=noNegZ(sin(phi));
  a21=noNegZ(-sin(phi)); a22=noNegZ(cos(phi));  
  printf("%s",print_float(a11,digits,buf)); printf("%s",print_float(a12,digits,buf));
  printf("  %5.2f   %5.2f",0.0,tr[0]);
  printf(" // rot(%s)  %s\n",print_frac(phi/M_PI,buf),comment);
  printf("%s",print_float(a21,digits,buf)); printf("%s",print_float(a22,digits,buf));
  printf("  %5.2f   %5.2f\n",0.0,tr[1]);
  printf("%s",print_float(0.0,digits,buf)); printf("%s",print_float(0.0,digits,buf));
  printf("  %5.2f   %5.2f\n",1.0,0.0);
}
  /* reflexion along axis with angle phi (positive) to x-axis 
     == rot(-phi)*ref(phi=0)*rot(phi) */
void ref(double phi,double *tr, char *comment)
{ double hcos, hsin; char buf[100]; double a11,a12,a21,a22;
  hcos=cos(phi); hsin=sin(phi);
  a11=noNegZ(hcos*hcos-hsin*hsin); a12=noNegZ(2.0*hcos*hsin);
  a21=noNegZ(2.0*hcos*hsin);       a22=noNegZ(hsin*hsin-hcos*hcos);
  printf("%s",print_float(a11,digits,buf)); printf("%s",print_float(a12,digits,buf));
  printf("  %5.2f   %5.2f",0.0,tr[0]);
  printf(" // ref(%s)  %s\n",print_frac(phi/M_PI,buf),comment);
  printf("%s",print_float(a21,digits,buf)); printf("%s",print_float(a22,digits,buf));
  printf("  %5.2f   %5.2f\n",0.0,tr[1]);
  printf("%s",print_float(0.0,digits,buf)); printf("%s",print_float(0.0,digits,buf));
  printf("  %5.2f   %5.2f\n",1.0,0.0);
}

  /* coefficient before the different terms of a spin Hamiltonian */
double term(int i, double phi) 
{  
    /* i=1..9: coefficient before the different terms of a
       scalar product of s_i and s_j, with an angle phi in between */
  switch (i) 
  {
    case  1: return( cos(phi) );            /* zz */
    case  2: return(-0.5*sin(phi));         /* z+ */
    case  3: return(-0.5*sin(phi));         /* z- */
    case  4: return( 0.5*sin(phi));         /* +z */
    case  5: return( 0.5*sin(phi));         /* -z */
    case  6: return( 0.25*(cos(phi)+1.0) ); /* +- */ 
    case  7: return( 0.25*(cos(phi)+1.0) ); /* -+ */
    case  8: return( 0.25*(cos(phi)-1.0) ); /* ++ */
    case  9: return( 0.25*(cos(phi)-1.0) ); /* -- */
      /* following terms for single spin 
         (usually only for an external field): */
    case 10: return(0.0);                   /* z  */
    case 11: return(0.0);                   /* -  */
    case 12: return(0.0);                   /* z^2 */
    case 13: return(0.0);                   /* +  */
    default: E0("invalid value i=%d",i);
  }
}

void print_term2(int i, int site, char *bonds, double factor, int dir)
{ int j, n_bonds; double h1; char buf[100];
/*  if ( (i>=10) && (i<=13) ) {
    printf("number=1 0.0\n");
    return;
  }*/
  n_bonds=0;
  for (j=0;j<n_nn;j++) {
    h1=bond[n_nn*(site-1)+j].value * term(i,bond[n_nn*(site-1)+j].angle);
    if ( (bonds[j]=='1') && (fabs(h1)>_nearZero) 
         && (directional*bond[n_nn*(site-1)+j].direct==dir) ) n_bonds++;
  }
  if (n_bonds==0) {
    printf("number=0\n");
  } else {
    printf("number=%d\n",n_bonds);
    for (j=0;j<n_nn;j++) {
      h1=factor * bond[n_nn*(site-1)+j].value * term(i,bond[n_nn*(site-1)+j].angle);
      if ( (bonds[j]=='1') && (fabs(h1)>_nearZero) 
           && (directional*bond[n_nn*(site-1)+j].direct==dir) ) {
        print_nn(site,j); 
        printf("%s\n",print_float(h1,digits,buf));
      }
    }
  }
}
  /* print the terms of Hamiltonian with comments */
void print_term(char *comment, int site, char *bonds, double factor)
{ int i;
  if (strlen(bonds)!=n_nn) E0("invalid strlen bonds:%s",bonds);
  for (i=1;i<=9;i++) {
    printf("// %s term %d %s\n",comment,i,term_descr[i]);
    if (directional==1) {
      print_term2(i,site,bonds,factor,1);
      print_term2(i,site,bonds,factor,-1);
    } else {  /* directional==0 */
      print_term2(i,site,bonds,factor,0);
    }    
  }
  for (i=10;i<=13;i++) {
    printf("// %s term %d %s\n",comment,i,term_descr[i]);
    printf("number=1 0.0\n");
  }
}
